/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#define FE_SDO_DEBUG	FALSE
#define FE_SDO_THREAD	FALSE
#define	FE_SDO_PROFILE	(FE_CODEGEN==FE_PROFILE)

#include "operator/operator.pmh"

namespace fe
{

void SubdivideOp::initialize(void)
{
#if !FE_SDO_THREAD
	catalog<bool>("Threads","hidden")=true;
	catalog<bool>("AutoThread","hidden")=true;
	catalog<bool>("Paging","hidden")=true;
	catalog<bool>("StayAlive","hidden")=true;
	catalog<bool>("Work","hidden")=true;
#endif

	catalog<String>("icon")="FE_alpha";

	catalog<String>("partitionAttr")="hull_primitive";
	catalog<String>("partitionAttr","label")="Partition Attr";
	catalog<String>("partitionAttr","hint")=
			"Primitive string attribute created on output"
			" distinguishing the original primitives.";

	catalog<bool>("recomputeNormals")=true;
	catalog<String>("recomputeNormals","label")="Recompute Normals";

	catalog<I32>("depth")=1;
	catalog<I32>("depth","high")=8;
	catalog<I32>("depth","max")=32;
	catalog<String>("depth","label")="Depth";
	catalog<String>("depth","hint")=
			"Number of subdivision iterations.";

	catalog<I32>("flatDepth")=0;
	catalog<I32>("flatDepth","high")=8;
	catalog<I32>("flatDepth","max")=32;
	catalog<String>("flatDepth","label")="Flat Depth";
	catalog<String>("flatDepth","hint")=
			"Number of iterations to linearly interpolate (at finest end).";

	catalog< sp<Component> >("Input Surface");

	//* don't copy input to output
	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";

	m_spSurfaceNormalOp=registry()->create("OperatorSurfaceI.SurfaceNormalOp");
}

void SubdivideOp::bind(sp<SignalerI> a_spSignalerI,sp<Layout> a_spLayout)
{
	OperatorSurfaceCommon::bind(a_spSignalerI,a_spLayout);

	sp<OperatorSurfaceCommon> spCommon=m_spSurfaceNormalOp;
	if(spCommon.isValid())
	{
		spCommon->bind(a_spSignalerI,a_spLayout);
	}
}

void SubdivideOp::clearData(void)
{
	m_facePoint.clear();
	m_faceSum.clear();
	m_faceCount.clear();
	m_midpointSum.clear();
	m_midpointCount.clear();
	m_boundarySum.clear();
	m_boundaryCount.clear();

	m_edgeMap.clear();
}

void SubdivideOp::handle(Record& a_rSignal)
{
#if FE_SDO_DEBUG
	feLog("SubdivideOp::handle\n");
#endif

#if !FE_SDO_THREAD
	catalog<bool>("AutoThread")=false;
	catalog<I32>("Threads")=1;
#endif

	m_spProfiler=new Profiler("SubdivideOp");
	m_spProfileAccess=new Profiler::Profile(m_spProfiler,"Access");
	m_spProfileResize=new Profiler::Profile(m_spProfiler,"Resize");
	m_spProfileSlot=new Profiler::Profile(m_spProfiler,"Slot");
	m_spProfileFace=new Profiler::Profile(m_spProfiler,"Face");
	m_spProfileFacePoint=new Profiler::Profile(m_spProfiler,"FacePoint");
	m_spProfileVertSize=new Profiler::Profile(m_spProfiler,"VertSize");
	m_spProfileEdgeMap=new Profiler::Profile(m_spProfiler,"EdgeMap");
	m_spProfileEdgePoint=new Profiler::Profile(m_spProfiler,"EdgePoint");
	m_spProfileVertex=new Profiler::Profile(m_spProfiler,"Vertex");
	m_spProfileAppend=new Profiler::Profile(m_spProfiler,"Append");
	m_spProfileEdge=new Profiler::Profile(m_spProfiler,"Edge");
	m_spProfilePoint=new Profiler::Profile(m_spProfiler,"Point");
	m_spProfileNormal=new Profiler::Profile(m_spProfiler,"Normal");

	m_spProfiler->begin();
	m_spProfileAccess->start();

	const String partitionAttr=catalog<String>("partitionAttr");
	const BWORD recomputeNormals=catalog<bool>("recomputeNormals");
	const I32 depth=catalog<I32>("depth");
	const I32 flatDepth=catalog<I32>("flatDepth");

	sp<SurfaceAccessibleI> spFinalOutputAccessible;
	if(!accessOutput(spFinalOutputAccessible,a_rSignal)) return;

	if(!access(m_spInputAccessible,"Input Surface")) return;

	if(depth<1)
	{
		spFinalOutputAccessible->copy(m_spInputAccessible);
	}

	for(I32 iteration=0;iteration<depth;iteration++)
	{
		m_flat=(iteration>=depth-flatDepth);

		//* shuffle buffers
		if(iteration)
		{
			m_spInputAccessible=m_spOutputAccessible;
		}

		if(iteration==depth-1)
		{
			m_spOutputAccessible=spFinalOutputAccessible;
		}
		else
		{
#if FALSE
			//* Record-based
			m_spOutputAccessible=registry()->create("SurfaceAccessibleI");
			sp<Scope> spScope=registry()->create("Scope");
			m_spOutputAccessible->bind(spScope);
#else
			m_spOutputAccessible=
					registry()->create("*.SurfaceAccessibleCatalog");
#endif
		}

		sp<SurfaceAccessorI> spOutputVertices;
		if(!access(spOutputVertices,m_spOutputAccessible,
				e_primitive,e_vertices)) return;

		sp<SurfaceAccessorI> spOutputPoint;
		if(!access(spOutputPoint,m_spOutputAccessible,
				e_point,e_position)) return;

		sp<SurfaceAccessorI> spOutputUV;
		if(!access(spOutputUV,m_spOutputAccessible,
				e_point,"uv")) return;

		sp<SurfaceAccessorI> spOutputPartition;
		if(!access(spOutputPartition,m_spOutputAccessible,
				e_primitive,partitionAttr)) return;

		sp<SurfaceAccessorI> spInputVertices;
		if(!access(spInputVertices,m_spInputAccessible,
				e_primitive,e_vertices)) return;

		sp<SurfaceAccessorI> spInputPoint;
		if(!access(spInputPoint,m_spInputAccessible,
				e_point,e_position)) return;

		sp<SurfaceAccessorI> spInputUV;
		if(!access(spInputUV,m_spInputAccessible,
				e_point,"uv")) return;

		sp<SurfaceAccessorI> spInputPartition;
		if(iteration && !access(spInputPartition,m_spInputAccessible,
				e_primitive,partitionAttr)) return;

		sp<SurfaceAccessorI> spOutputNormal;
		sp<SurfaceAccessorI> spInputNormal;
		if(!recomputeNormals)
		{
			if(!access(spOutputNormal,m_spOutputAccessible,
					e_point,e_normal)) return;

			if(!access(spInputNormal,m_spInputAccessible,
					e_point,e_normal)) return;
		}

		//* Catmull-Clark

		const U32 primitiveCountIn=spInputVertices->count();
		const U32 pointCountIn=spInputPoint->count();

#if FE_SDO_DEBUG
		feLog("SubdivideOp::handle input points %d primitives %d\n",
				pointCountIn,primitiveCountIn);
#endif

		m_spProfileResize->replace(m_spProfileAccess);

		clearData();

		m_facePoint.resize(primitiveCountIn);

		m_faceSum.resize(pointCountIn,SpatialVector(0.0,0.0,0.0));

		m_faceCount.resize(pointCountIn,0);

		m_midpointSum.resize(pointCountIn,SpatialVector(0.0,0.0,0.0));

		m_midpointCount.resize(pointCountIn,0);

		m_boundarySum.resize(pointCountIn,SpatialVector(0.0,0.0,0.0));

		m_boundaryCount.resize(pointCountIn,0);

		m_spProfileSlot->replace(m_spProfileResize);

		SpatialVector pointsIn[pointCountIn];
		SpatialVector uvsIn[pointCountIn];
		SpatialVector normalsIn[recomputeNormals? 1: pointCountIn];

		m_pointsIn=pointsIn;
		m_uvsIn=uvsIn;
		m_normalsIn=normalsIn;

		spOutputPoint->append(pointCountIn+primitiveCountIn);

		//* make slots for original points (with adjustments)
		for(U32 pointIndexIn=0;pointIndexIn<pointCountIn;pointIndexIn++)
		{
			m_pointsIn[pointIndexIn]=spInputPoint->spatialVector(pointIndexIn);

			SpatialVector& rUv=m_uvsIn[pointIndexIn];
			rUv=spInputUV->spatialVector(pointIndexIn);
			spOutputUV->set(pointIndexIn,rUv);

			if(!recomputeNormals)
			{
				SpatialVector& rNormal=m_normalsIn[pointIndexIn];
				rNormal=spInputNormal->spatialVector(pointIndexIn);
				spOutputNormal->set(pointIndexIn,rNormal);
			}
		}

		U32 primVertInc=4096;
		U32 primVertSize=primitiveCountIn*5;	//* probably overshoot
		std::vector< std::vector<I32> > primVerts(primVertSize);
		U32 primVertCount=0;

		m_spProfileFace->replace(m_spProfileSlot);

		for(U32 primitiveIndexIn=0;primitiveIndexIn<primitiveCountIn;
				primitiveIndexIn++)
		{
			String partition;
			if(iteration)
			{
				partition=spInputPartition->string(primitiveIndexIn);
			}
			else
			{
				partition.sPrintf("%d",primitiveIndexIn);
			}

			SpatialVector pointSum(0.0,0.0,0.0);
			SpatialVector normalSum(0.0,0.0,0.0);
			SpatialVector uvSum(0.0,0.0,0.0);

			//* UV wrap-around
			Vector2i highUV(false,false);
			SpatialVector lowUV(0.0,0.0,0.0);

			const U32 subCount=spInputVertices->subCount(primitiveIndexIn);

			I32 pointIndices[subCount];

			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const I32 pointIndexIn=
						spInputVertices->integer(primitiveIndexIn,subIndex);
				pointIndices[subIndex]=pointIndexIn;

				const SpatialVector uv=m_uvsIn[pointIndexIn];
				pointSum+=m_pointsIn[pointIndexIn];
				uvSum+=uv;

				if(uv[0]>0.7)
				{
					highUV[0]=true;
				}
				if(uv[1]>0.7)
				{
					highUV[1]=true;
				}
				lowUV[0]+=(uv[0]<0.3);
				lowUV[1]+=(uv[1]<0.3);

				if(!recomputeNormals)
				{
					normalSum+=m_normalsIn[pointIndexIn];
				}
			}

			if(highUV[0])
			{
				uvSum[0]+=lowUV[0];
			}
			if(highUV[1])
			{
				uvSum[1]+=lowUV[1];
			}

			if(subCount)
			{
				const Real scalar=1.0/Real(subCount);

				pointSum*=scalar;
				uvSum*=scalar;

				if(!recomputeNormals)
				{
					normalSum=unitSafe(normalSum);
				}
			}

			if(uvSum[0]>=1.0)
			{
				uvSum[0]-=1.0;
			}
			if(uvSum[0]>=1.0)
			{
				uvSum[0]-=1.0;
			}

			m_spProfileFacePoint->replace(m_spProfileFace);

			//* face point
//			const I32 pointIndexOut=spOutputPoint->append();
			const I32 pointIndexOut=pointCountIn+primitiveIndexIn;

			spOutputPoint->set(pointIndexOut,pointSum);
			spOutputUV->set(pointIndexOut,uvSum);

			if(!recomputeNormals)
			{
				spOutputNormal->set(pointIndexOut,normalSum);
			}

			m_facePoint[primitiveIndexIn]=pointIndexOut;

			if(primVertCount+subCount>=primVertSize)
			{
				m_spProfileVertSize->replace(m_spProfileFacePoint);

				primVertSize+=primVertInc;
				primVerts.resize(primVertSize);

				m_spProfileFacePoint->replace(m_spProfileVertSize);
			}

			m_spProfileEdgeMap->replace(m_spProfileFacePoint);

			I32 wrapIndex= -1;
			I32 lastEdgePointIndex= -1;
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				//* edge face
				const I32 pointIndexA=pointIndices[subIndex];
				const I32 pointIndexB=pointIndices[(subIndex+1)%subCount];

				Edge& rEdge=(pointIndexA>pointIndexB)?
						m_edgeMap[(U64(pointIndexA)<<32)+pointIndexB]:
						m_edgeMap[(U64(pointIndexB)<<32)+pointIndexA];

				rEdge.m_faceSet.insert(primitiveIndexIn);

				m_spProfileEdgePoint->replace(m_spProfileEdgeMap);

				if(rEdge.m_pointIndex<0)
				{
					rEdge.m_pointIndex=spOutputPoint->append();
				}

				m_spProfileVertex->replace(m_spProfileEdgePoint);

				//* face influence on pre-existing points
				m_faceSum[pointIndexA]+=pointSum;
				m_faceCount[pointIndexA]++;

				//* create sub primitive
#if FALSE
				const I32 primitiveIndexOut=spOutputVertices->append();

				spOutputPartition->set(primitiveIndexOut,partition);

				//* vertex 0: original point
				spOutputVertices->append(primitiveIndexOut,pointIndexA);

				//* vertex 1: midpoint
				spOutputVertices->append(primitiveIndexOut,rEdge.m_pointIndex);

				//* vertex 2: face point
				spOutputVertices->append(primitiveIndexOut,pointIndexOut);

				//* vertex 3: prev midpoint
				spOutputVertices->append(primitiveIndexOut,lastEdgePointIndex);

				if(subIndex==subCount-1)
				{
					spOutputVertices->set(wrapIndex,3,rEdge.m_pointIndex);
				}
				if(!subIndex)
				{
					wrapIndex=primitiveIndexOut;
				}
#else
				std::vector<I32>& rVerts=primVerts[primVertCount++];
				rVerts.resize(4);

//				const I32 primitiveIndexOut=spOutputVertices->count();

				//* TODO
//				spOutputPartition->set(primitiveIndexOut,partition);

				//* vertex 0: original point
				rVerts[0]=pointIndexA;

				//* vertex 1: midpoint
				rVerts[1]=rEdge.m_pointIndex;

				//* vertex 2: face point
				rVerts[2]=pointIndexOut;

				//* vertex 3: prev midpoint
				rVerts[3]=lastEdgePointIndex;

				if(subIndex==subCount-1)
				{
					primVerts[wrapIndex][3]=rEdge.m_pointIndex;
				}
				if(!subIndex)
				{
					wrapIndex=primVertCount-1;
				}
#endif

				lastEdgePointIndex=rEdge.m_pointIndex;

				m_spProfileEdgeMap->replace(m_spProfileVertex);
			}

			m_spProfileFace->replace(m_spProfileEdgeMap);
		}

		primVerts.resize(primVertCount);

		m_spProfileAppend->replace(m_spProfileFace);

		spOutputVertices->append(primVerts);

		m_spProfileEdge->replace(m_spProfileAppend);

		//* set edge points
		adjustThreads();	//* in case auto was just turned on
		const U32 threadCount=catalog<I32>("Threads");
		setNonAtomicRange(0,threadCount-1);
		setStage("edge");
		OperatorThreaded::handle(a_rSignal);

		m_spProfilePoint->replace(m_spProfileEdge);

		//* adjust original points
		setAtomicRange(m_spInputAccessible,e_pointsOnly);
		setStage("point");
		OperatorThreaded::handle(a_rSignal);

		m_spProfileAccess->replace(m_spProfilePoint);

#if FE_SDO_DEBUG
		feLog("SubdivideOp::handle"
				" iteration %d subdivided points %d primitives %d\n",
				iteration,spOutputPoint->count(),spOutputVertices->count());
#endif
	}

	if(recomputeNormals)
	{
		m_spProfileNormal->replace(m_spProfileAccess);

		sp<Catalog> spOtherCatalog=m_spSurfaceNormalOp;
		if(spOtherCatalog.isValid())
		{
			spOtherCatalog->catalog<I32>("Threads")=
					catalog<I32>("Threads");
			spOtherCatalog->catalog<bool>("AutoThread")=
					catalog<bool>("AutoThread");
			spOtherCatalog->catalog<bool>("Paging")=
					catalog<bool>("Paging");
			spOtherCatalog->catalog<bool>("StayAlive")=
					catalog<bool>("StayAlive");
		}

		m_spSurfaceNormalOp->handle(a_rSignal);

		m_spProfileAccess->replace(m_spProfileNormal);
	}

	clearData();

	m_spProfileAccess->finish();
	m_spProfiler->end();

#if FE_SDO_PROFILE
	feLog("%s:\n%s\n",m_spProfiler->name().c_str(),
			m_spProfiler->report().c_str());
#endif
}

void SubdivideOp::run(I32 a_id,sp<SpannedRange> a_spRange,String a_stage)
{
#if FE_SDO_DEBUG
	feLog("SubdivideOp::run id %d stage \"%s\" range %s\n",
			a_id,a_stage.c_str(),a_spRange->brief().c_str());
#endif

	if(a_stage=="edge")
	{
		run_edge(a_id,a_spRange);
	}
	else if(a_stage=="point")
	{
		run_point(a_id,a_spRange);
	}
}

void SubdivideOp::run_edge(I32 a_id,sp<SpannedRange> a_spRange)
{
#if FE_SDO_DEBUG
	feLog("SubdivideOp::run_edge id %d\n",a_id);
#endif

	const BWORD subdivideNormals=!catalog<bool>("recomputeNormals");

	sp<SurfaceAccessorI> spOutputPoint;
	if(!access(spOutputPoint,m_spOutputAccessible,e_point,e_position)) return;

	sp<SurfaceAccessorI> spOutputNormal;
	if(subdivideNormals &&
			!access(spOutputNormal,m_spOutputAccessible,
			e_point,e_normal)) return;

	sp<SurfaceAccessorI> spOutputUV;
	if(!access(spOutputUV,m_spOutputAccessible,e_point,"uv")) return;

	sp<SurfaceAccessorI> spInputPoint;
	if(!access(spInputPoint,m_spInputAccessible,e_point,e_position)) return;

	//* assuming contiguous
	SpannedRange::Iterator it=a_spRange->begin();
	if(it.atEnd())
	{
		return;
	}

	const U32 modulus=it.value();

#if FE_SDO_THREAD
	const U32 pointCountIn=spInputPoint->count();

	const U32 threadCount=catalog<I32>("Threads");
	const I32 pointMin=modulus*pointCountIn/threadCount;
	const I32 pointMax=(modulus==threadCount-1)? pointCountIn-1:
			(modulus+1)*pointCountIn/threadCount-1;

//	feLog("mod %d/%d count %d span %d %d\n",
//			modulus,threadCount,pointCountIn,pointMin,pointMax);

#else
	//* all on one thread
	if(modulus)
	{
		return;
	}
//	const I32 pointMin=0;
//	const I32 pointMax=pointCountIn-1;
#endif

	for(std::map<U64,Edge>::iterator it=m_edgeMap.begin();
			it!=m_edgeMap.end();it++)
	{
		const U64 combo=it->first;
		const Edge& rEdge=it->second;

		const I32 pointIndexA=combo>>32;
		const I32 pointIndexB=combo&0xFFFFFFFF;

#if FE_SDO_THREAD
		const BWORD changeA=(pointIndexA>=pointMin && pointIndexA<=pointMax);
		const BWORD changeB=(pointIndexB>=pointMin && pointIndexB<=pointMax);

		if(!changeA && !changeB)
		{
			continue;
		}
#endif

		const std::set<I32>& rFaceSet=rEdge.m_faceSet;

		const SpatialVector pointA=m_pointsIn[pointIndexA];
		const SpatialVector uvA=m_uvsIn[pointIndexA];

		const SpatialVector pointB=m_pointsIn[pointIndexB];
		const SpatialVector uvB=m_uvsIn[pointIndexB];

		SpatialVector pointSum=pointA+pointB;
		SpatialVector uvSum=uvA+uvB;

		SpatialVector normalA;
		SpatialVector normalB;
		SpatialVector normalSum;
		if(subdivideNormals)
		{
			normalA=m_normalsIn[pointIndexA];
			normalB=m_normalsIn[pointIndexB];
			normalSum=normalA+normalB;
		}

		const SpatialVector midpoint=0.5*pointSum;
		I32 divisor=2;

		//* UV wrap-around
		Vector2i highUV(
				(uvA[0]>0.7 || uvB[0]>0.7),
				(uvA[1]>0.7 || uvB[1]>0.7));
		Vector2i lowUV(
				(uvA[0]<0.3)+(uvB[0]<0.3),
				(uvA[1]<0.3)+(uvB[1]<0.3));

		if(!m_flat && rFaceSet.size()>1)
		{
			//* also add adjacent face points
			for(std::set<I32>::iterator it=rFaceSet.begin();
					it!=rFaceSet.end();it++)
			{
				const I32 primitiveIndexIn=*it;
				const I32 pointIndexOut=m_facePoint[primitiveIndexIn];
				const SpatialVector uv=
						spOutputUV->spatialVector(pointIndexOut);

				pointSum+=spOutputPoint->spatialVector(pointIndexOut);
				uvSum+=uv;
				divisor++;

				if(subdivideNormals)
				{
					normalSum+=spOutputNormal->spatialVector(pointIndexOut);
				}

				if(uv[0]>0.7)
				{
					highUV[0]=TRUE;
				}
				if(uv[1]>0.7)
				{
					highUV[1]=TRUE;
				}
				lowUV[0]+=(uv[0]<0.3);
				lowUV[1]+=(uv[1]<0.3);
			}

#if FE_SDO_THREAD
			if(changeA)
#endif
			{
				m_midpointSum[pointIndexA]+=midpoint;
				m_midpointCount[pointIndexA]++;
			}

#if FE_SDO_THREAD
			if(changeB)
#endif
			{
				m_midpointSum[pointIndexB]+=midpoint;
				m_midpointCount[pointIndexB]++;
			}
		}
		else
		{
#if FE_SDO_THREAD
			if(changeA)
#endif
			{
				m_boundarySum[pointIndexA]+=midpoint;
				m_boundaryCount[pointIndexA]++;
			}

#if FE_SDO_THREAD
			if(changeB)
#endif
			{
				m_boundarySum[pointIndexB]+=midpoint;
				m_boundaryCount[pointIndexB]++;
			}
		}

		//* NOTE rEdge.m_pointIndex isn't pointIndexA,
		//* but this should keep it exclusive
#if FE_SDO_THREAD
		if(!changeA)
		{
			continue;
		}
#endif

		if(highUV[0])
		{
			uvSum[0]+=Real(lowUV[0]);
		}
		if(highUV[1])
		{
			uvSum[1]+=Real(lowUV[1]);
		}

		const Real scalar=1.0/Real(divisor);

		pointSum*=scalar;
		uvSum*=scalar;

		if(uvSum[0]>=1.0)
		{
			uvSum[0]-=1.0;
		}
		if(uvSum[0]>=1.0)
		{
			uvSum[0]-=1.0;
		}

		spOutputPoint->set(rEdge.m_pointIndex,pointSum);
		spOutputUV->set(rEdge.m_pointIndex,uvSum);

		if(subdivideNormals)
		{
			normalSum=unitSafe(normalSum);

			spOutputNormal->set(rEdge.m_pointIndex,normalSum);
		}
	}
}

void SubdivideOp::run_point(I32 a_id,sp<SpannedRange> a_spRange)
{
#if FE_SDO_DEBUG
	feLog("SubdivideOp::run_point id %d\n",a_id);
#endif

	sp<SurfaceAccessorI> spOutputPoint;
	if(!access(spOutputPoint,m_spOutputAccessible,
			e_point,e_position)) return;

	I32 faults=0;

	if(m_flat)
	{
		for(SpannedRange::Iterator it=a_spRange->begin();!it.atEnd();it.step())
		{
			const U32 pointIndexIn=it.value();

			spOutputPoint->set(pointIndexIn,m_pointsIn[pointIndexIn]);
		}
	}
	else
	{
		for(SpannedRange::Iterator it=a_spRange->begin();!it.atEnd();it.step())
		{
			const U32 pointIndexIn=it.value();

			const SpatialVector pointIn=m_pointsIn[pointIndexIn];

			const I32 pointFaceCount=m_faceCount[pointIndexIn];
			FEASSERT(pointFaceCount>0);
			if(pointFaceCount<1)
			{
					String message;
					message.sPrintf("pointIndex %d has zero adjacent faces;",
							pointIndexIn);
					catalog<String>("warning")+=message;

				continue;
			}

			const I32 pointBoundaryCount=m_boundaryCount[pointIndexIn];

			const SpatialVector faceAverage=
					m_faceSum[pointIndexIn]/Real(pointFaceCount);

			if(pointFaceCount>2 || pointBoundaryCount<1)
			{
				if(pointFaceCount<3)
				{
					FEASSERT(!pointBoundaryCount);

					String message;
					message.sPrintf("pointIndex %d is used on only %d %s,"
							" but has no adjacent boundary edges;",
							pointIndexIn,pointFaceCount,
							pointFaceCount==1? "face": "faces");
					catalog<String>("warning")+=message;
				}

				const I32 divisor=m_midpointCount[pointIndexIn]+
						pointBoundaryCount;
				FEASSERT(divisor>0);
				if(divisor<1)
				{
					String message;
					message.sPrintf("pointIndex %d has a zero divisor;",
							pointIndexIn);
					catalog<String>("error")+=message;

					faults++;
					continue;
				}

				const SpatialVector midpointAverage=
						(m_midpointSum[pointIndexIn]+
						m_boundarySum[pointIndexIn])/Real(divisor);

				const SpatialVector pointOut=
						(faceAverage+2.0*midpointAverage+
						(pointFaceCount-3)*pointIn)/Real(pointFaceCount);

				spOutputPoint->set(pointIndexIn,pointOut);
			}
			else
			{
				//* TODO find out what everyone else does on boundaries

				FEASSERT(pointBoundaryCount>0);

				const SpatialVector boundaryAverage=
						m_boundarySum[pointIndexIn]/Real(pointBoundaryCount);

				const SpatialVector pointOut=
						0.5*(boundaryAverage+pointIn);

				spOutputPoint->set(pointIndexIn,pointOut);
			}
		}
	}

	if(faults)
	{
		feX("SubdivideOp::run_point","input mesh is non-compliant");
	}
}

} // namespace
